"""*******************************************************************************
* ACETONE: Predictable programming framework for ML applications in safety-critical systems
* Copyright (c) 2022. ONERA
* This file is part of ACETONE
*
* ACETONE is free software ;
* you can redistribute it and/or modify it under the terms of the GNU Lesser General Public
* License as published by the Free Software Foundation ;
* either version 3 of  the License, or (at your option) any later version.
*
* ACETONE is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY ;
* without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
* See the GNU Lesser General Public License for more details.
*
* You should have received a copy of the GNU Lesser General Public License along with this program ;
* if not, write to the Free Software Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307  USA
******************************************************************************
"""

from abc import ABC
from pathlib import Path
from typing import Any

import numpy as np
import onnx
import pystache
from keras.engine.functional import Functional
from keras.engine.sequential import Sequential

from acetone_nnet import templates
from acetone_nnet.code_generator.layers import (
    AveragePooling2D,
    BatchNormalization,
    Broadcast,
    Concatenate,
    Conv2D,
    Conv2D_6loops,
    Conv2D_indirect_gemm,
    Conv2D_std_gemm,
    Dense,
    Dot,
    Gather,
    GatherElements,
    Gemm,
    MatMul,
    MaxPooling2D,
    Pad,
    Pooling2D,
    Reduce,
    ResizeCubic,
    ResizeLinear,
    ResizeNearest,
    Softmax,
)
from acetone_nnet.format_importer.parser import parser
from acetone_nnet.versioning.versioning import versioning

class CodeGenerator(ABC):

    def __init__(
        self,
        file: str | Path | onnx.ModelProto | Functional | Sequential,
        test_dataset: str | np.ndarray | None = None,
        function_name: str = "inference",
        nb_tests: int | str = 0,
        conv_algorithm: str = "std_gemm_nn",
        normalize: bool | str = False,
        debug_mode: str | None = None,
        **kwargs,
    ):

        self.file = file
        self.function_name = function_name
        self.nb_tests = int(nb_tests)
        self.conv_algorithm = conv_algorithm
        self.normalize = bool(normalize)
        self.template_path = templates.__file__[:-11]

        if not self.normalize:
            l, dtype, dtype_py, data_format, maxpath, dict_cst = parser(
                self.file,
                self.conv_algorithm,
            )
        else:  # normalise
            l, dtype, dtype_py, data_format, maxpath, dict_cst, Normalizer = parser(
                self.file,
                self.conv_algorithm,
                normalize=self.normalize,
            )
            self.Normalizer = Normalizer

        self.layers: list[Any] = l
        version = {}
        for i in range(len(self.layers)):
            if self.layers[i].name == "Conv2D":
                version[i] = "indirect_gemm_nn"
        
        self.layers = versioning(self.layers, l)
            
        self.data_type = dtype
        self.data_type_py = dtype_py
        self.maxpath = maxpath
        self.data_format = data_format
        self.dict_cst = dict_cst

        if isinstance(test_dataset, str):
            self.test_dataset_file = test_dataset
            ds = self.load_test_dataset()
            self.test_dataset = ds
        elif isinstance(test_dataset, np.ndarray):
            self.test_dataset = test_dataset
        else:
            print("creating random dataset")
            ds = self.create_test_dataset()
            self.test_dataset = ds

        self.files_to_gen = [
            "inference.c",
            "inference.h",
            "global_vars.c",
            "main.c",
            "Makefile",
            "test_dataset.h",
            "test_dataset.c",
        ]

        ##### Debug Mode #####
        self.debug_mode = debug_mode
        if self.debug_mode:
            self.debug_target = self.load_debug_target(debug_mode)
        ##### Debug Mode #####

        ####### Checking the instantiation#######

        ### Checking argument type ###
        if not isinstance(
            self.file,
            str | Path | onnx.ModelProto | Functional | Sequential,
        ):
            raise TypeError(
                "Error: model type.\n Format must be: path to model, model ONNX or model Keras",
            )
        if not (isinstance(test_dataset, str | np.ndarray) or test_dataset is None):
            raise TypeError(
                "Error: test_dataset type.\n Must be: path to dataset text filen, numpy array or None",
            )
        if not isinstance(self.function_name, str):
            raise TypeError("Error: function name type.\n Must be a string")
        if not isinstance(self.conv_algorithm, str):
            raise TypeError("Error: conv algorihtm type.\n Must be a string")
        if not (self.debug_mode is None or isinstance(self.debug_mode, str)):
            raise TypeError("Error: debug mode type.\n Must be: string or None")
        if self.debug_mode:
            assert isinstance(self.debug_target, list)

        ### Checking value consistency ###
        if self.conv_algorithm not in [
            "6loops",
            "indirect_gemm_nn",
            "indirect_gemm_tn",
            "indirect_gemm_nt",
            "indirect_gemm_tt",
            "std_gemm_nn",
            "std_gemm_tn",
            "std_gemm_nt",
            "std_gemm_tt",
        ]:
            raise ValueError(
                "Error: conv algorithm value.\n Must be one of: 6loops, indirect_gemm_nn, indirect_gemm_tn, indirect_gemm_nt, indirect_gemm_tt, std_gemm_nn, std_gemm_tn, std_gemm_nt, std_gemm_tt",
            )

        # Debug Mode
        if self.debug_mode and self.debug_mode not in ["keras", "onnx"]:
            raise ValueError(
                "Error: debug mode value.\n Must be one of: keras, onnx",
            )

    def load_debug_target(self, debug_mode: str | None) -> list[int]:
        """Identify list of layers indices to debug."""
        targets: list[int] = []
        for layer in self.layers[1:]:
            if debug_mode == "keras" and layer.name == "Softmax":
                targets[-1] = layer.idx
            else:
                targets.append(layer.idx)

        return targets

    def create_test_dataset(self) -> np.ndarray:
        """Create random dataset for graph."""
        return np.random.default_rng(seed=10).random(
            size=(self.nb_tests, 1, int(self.layers[0].size)),
            dtype=self.data_type_py,
        )

    def load_test_dataset(self) -> np.ndarray:
        """Load test dataset from file."""
        test_dataset: list[list[int | float | np.float32]] = []
        with Path(self.test_dataset_file).open() as f:
            for i, line in enumerate(f):
                contents = line[1:-2].split(",")
                if self.data_type == "int":
                    test_dataset.append(list(map(int, contents)))
                elif self.data_type == "double":
                    test_dataset.append(list(map(float, contents)))
                elif self.data_type == "float":
                    test_dataset.append(list(map(np.float32, contents)))
                if i >= self.nb_tests - 1:
                    break
        return np.array(test_dataset)

    def compute_inference(
        self,
        c_files_directory: str,
    ) -> tuple[list, list] | np.ndarray:
        """Perform inference pass on test dataset."""
        with (Path(c_files_directory) / "output_python.txt").open("w+") as fi:
            for nn_input in self.test_dataset:

                # Debug Mode output
                debug_output: list[np.ndarray] = []
                targets: list[str] = []

                # Prepare graph input
                if (self.data_format == "channels_last") and (
                    len(self.layers[0].input_shape) == 4
                ):
                    shape = (
                        self.layers[0].input_shape[2],
                        self.layers[0].input_shape[3],
                        self.layers[0].input_shape[1],
                    )
                    nn_input = np.transpose(np.reshape(nn_input, shape), (2, 0, 1))
                if self.normalize:
                    nn_input = self.Normalizer.pre_processing(nn_input)

                # Inputs per layer and predecessor
                # - li[i][j] is an input for i computed from preceding layer j
                # - li[i][i] is the input for layers with no predecessors
                layer_inputs: dict[int, dict[int, np.ndarray]] = {
                    i.idx: {i.idx: nn_input} for i in self.layers
                }

                def gather_inputs(
                    target,
                    target_inputs,
                ) -> np.ndarray | list[np.ndarray]:
                    # Prepare inputs for computation
                    inputs: np.ndarray | list[np.ndarray] = []
                    if not target.previous_layer:
                        inputs.append(target_inputs[target.idx][target.idx])
                    else:
                        inputs.extend(
                            [
                                target_inputs[target.idx][p.idx]
                                for p in target.previous_layer
                            ],
                        )
                    return inputs if len(inputs) > 1 else inputs[0]

                # Forward computation layer by layer
                #   (Assumes layers are topologically sorted)
                for layer in self.layers:
                    # Ensure all inputs of layer are ready
                    if any(
                        p.idx not in layer_inputs[layer.idx]
                        for p in layer.previous_layer
                    ):
                        raise NotImplementedError

                    # Prepare inputs for computation
                    forward_inputs = gather_inputs(layer, layer_inputs)

                    # Compute layer output
                    layer_output = layer.forward_path_layer(forward_inputs)
                    if layer.fused_layer:
                        fused_inputs = gather_inputs(layer.fused_layer, layer_inputs)
                        layer_output = layer.fused_layer.forward_path_layer(
                            fused_inputs,
                        )

                    # Write Layer output into successors' input
                    for successor in layer.next_layer:
                        layer_inputs[successor.idx][layer.idx] = layer_output

                    # Cleanup layer inputs
                    del layer_inputs[layer.idx]

                    # Debug Mode
                    if self.debug_mode and layer.name != "Input_layer":
                        # Add the inference result of the layer to debug_output
                        if layer.idx in self.debug_target:
                            debug_output.append(layer_output)
                            if (self.data_format == "channels_last") and hasattr(
                                layer,
                                "output_channels",
                            ):
                                debug_output[-1] = np.transpose(
                                    debug_output[-1],
                                    (1, 2, 0),
                                )
                            debug_output[-1] = debug_output[-1].flatten()

                        targets.append(str(layer.name) + " " + str(layer.idx))

                nn_output = layer_output

                # Write results in text files to compare prediction.
                if (self.data_format == "channels_last") and hasattr(
                    layer,
                    "output_channels",
                ):
                    nn_output = np.transpose(nn_output, (1, 2, 0))
                nn_output = np.reshape(nn_output, -1)

                if self.normalize:
                    nn_output = self.Normalizer.post_processing(nn_output)

                for n in nn_output:
                    print(f"{n:.9g}", end=" ", file=fi, flush=True)
                print(" ", file=fi)

        print("File output_python.txt generated.")

        if self.debug_mode:
            return debug_output, targets
        return nn_output

    @staticmethod
    def flatten_array_orderc(array: np.ndarray) -> str:
        """Generate C flat array initializer in C order."""
        flattened_aray = array.flatten(order="C")
        s = "\n        {"
        for i in range(flattened_aray.size):
            s += str(flattened_aray[i]) + ", "
        s = s[:-2]
        s += "}"

        return s

    @staticmethod
    def flatten_array(array: np.ndarray) -> str:
        """Generate C flat array initializer in Fortran order."""
        s = "\n        {"
        shape = array.shape
        if len(shape) < 4:
            for i in range(4 - len(shape)):
                shape = (1,) + shape
            array = np.reshape(array, shape)
        for j in range(shape[3]):
            for k in range(shape[0]):
                for f in range(shape[1]):
                    for i in range(shape[2]):
                        s += str(array[k, f, i, j]) + ", "
        s = s[:-2]
        s += "}"
        return s

    def generate_testdataset_files(self, output_dir: Path):
        """Generate C code for graph test dataset."""
        # Generate test header file
        template = (
            Path(self.template_path) / "template_test_dataset_header.c.tpl"
        ).read_text()
        with open(output_dir / "test_dataset.h", "w+") as testdataset_header:
            testdataset_header.write(
                pystache.render(
                    template,
                    {
                        "nb_tests": self.nb_tests,
                        "nb_inputs": self.layers[0].size,
                        "nb_outputs": self.layers[-1].size,
                        "data_type": self.data_type,
                    },
                ),
            )

        # Generate test source file
        dataset = "{"
        if self.test_dataset is not None:
            dataset += ",".join(
                map(CodeGenerator.flatten_array_orderc, self.test_dataset),
            )
        dataset += "};\n"

        template = (
            Path(self.template_path) / "template_test_dataset_source.c.tpl"
        ).read_text()
        with open(output_dir / "test_dataset.c", "w+") as testdataset_source:
            testdataset_source.write(
                pystache.render(
                    template,
                    {"data_type": self.data_type, "dataset": dataset},
                ),
            )

    def generate_main_file(self, output_dir: Path):
        """Generate entry point C code."""
        template = (Path(self.template_path) / "template_main_file.c.tpl").read_text()
        with (output_dir / "main.c").open("a+") as main_file:
            main_file.write(pystache.render(template, {"data_type": self.data_type}))

    def generate_makefile(self, output_dir: Path):
        """Generate Makefile build script."""
        header_files = []
        source_files = []
        for filename in self.files_to_gen:
            if ".c" in filename:
                source_files.append(filename)
            elif ".h" in filename:
                header_files.append(filename)

        template = (Path(self.template_path) / "template_Makefile.tpl").read_text()

        with (output_dir / "Makefile").open("a+") as makefile:
            makefile.write(
                pystache.render(
                    template,
                    {
                        "source_files": " ".join(source_files),
                        "header_files": " ".join(header_files),
                        "function_name": self.function_name,
                    },
                ),
            )

    def generate_c_files(self, c_files_directory: str | Path) -> None:
        """Generate C code implementation of current graph."""
        # Prepare output directory
        c_files_directory = Path(c_files_directory)
        c_files_directory.mkdir(exist_ok=True, parents=True)

        for file in self.files_to_gen:
            if (c_files_directory / file).exists():
                raise FileExistsError(c_files_directory / file)

        self.generate_function_source_file(c_files_directory)
        print("Generated function source file.")
        self.generate_function_header_file(c_files_directory)
        print("Generated function header file.")
        self.generate_globalvars_file(c_files_directory)
        print("Generated globalvars .c file.")
        self.generate_main_file(c_files_directory)
        print("Generated main file.")
        self.generate_makefile(c_files_directory)
        print("Generated Makefile.")
        self.generate_testdataset_files(c_files_directory)
        print("Generated testdataset files.")

    def generate_function_source_file(self, output_dir: Path):
        """Generate C Code for inference function."""
        mustach_hash = {
            "data_type": self.data_type,
            "input_size": self.layers[0].size,
            "output_size": self.layers[-1].size,
        }

        # Tag Gather-type layers
        if any(isinstance(i, Gather | GatherElements) for i in self.layers):
            gather_layers = [
                i for i in self.layers if isinstance(i, Gather | GatherElements)
            ]
            mustach_hash["is_gather"] = True
            indices = []
            for gather in gather_layers:
                indices.append(
                    {
                        "idx": f"{gather.idx:02d}",
                        "lenght": len(gather.indices.flatten()),
                        "list": self.flatten_array_orderc(gather.indices),
                    },
                )
            mustach_hash["indices"] = indices

        self.l_size_max = max((i.size for i in self.layers), default=1)

        if any(isinstance(i, Dot | Pooling2D | Conv2D | Gemm) for i in self.layers):
            mustach_hash["p"] = True

        if any(
            isinstance(i, Conv2D_6loops | Conv2D_std_gemm | Pooling2D | Gemm)
            for i in self.layers
        ):
            mustach_hash["hw"] = True

        if any(isinstance(i, Dense | MatMul) for i in self.layers):
            mustach_hash["is_dense"] = True

        if any(
            isinstance(i, Conv2D_6loops | AveragePooling2D | Softmax)
            for i in self.layers
        ):
            mustach_hash["is_sum"] = True

        if any(isinstance(i, MaxPooling2D) for i in self.layers):
            mustach_hash["is_max"] = True

        if any(isinstance(i, AveragePooling2D) for i in self.layers):
            mustach_hash["is_count"] = True

        if any(
            isinstance(i, ResizeLinear | ResizeCubic | ResizeNearest)
            for i in self.layers
        ):
            mustach_hash["is_resize"] = True

        if any(isinstance(layer, ResizeCubic) for layer in self.layers):
            mustach_hash["is_cubic_interpolation"] = True

        if any(isinstance(layer, ResizeLinear) for layer in self.layers):
            mustach_hash["is_linear_interpolation"] = True

        if any(isinstance(layer, Reduce) for layer in self.layers):
            mustach_hash["is_reduced"] = True

        if self.debug_mode:
            mustach_hash["debug_file"] = output_dir / "debug_file.txt"

        # Generate parameters per layer
        mustach_hash["layers"] = []
        for layer in self.layers:
            layer_hash = {
                "inference_function": layer.generate_inference_code_layer(),
                "path": layer.path,
                "size": layer.size,
            }

            if self.dict_cst and layer in self.dict_cst:
                layer_hash["cst"] = True
                layer_hash["cst_name"] = self.dict_cst[layer.idx]

            if self.debug_mode and layer.idx in self.debug_target:
                layer_hash["debug_layer"] = True
                layer_hash["name"] = layer.name
                layer_hash["idx"] = layer.idx
                layer_hash["to_transpose"] = 0
                if (self.data_format == "channels_last") and (
                    hasattr(layer, "output_channels")
                ):
                    layer_hash["to_transpose"] = 1
                    layer_hash["channels"] = layer.output_channels
                    layer_hash["height"] = layer.output_height
                    layer_hash["width"] = layer.output_width

            mustach_hash["layers"].append(layer_hash)

        # Generate code to output graph data
        output_hash = {"path": self.layers[-1].path}
        if (self.data_format == "channels_last") and (
            hasattr(self.layers[-1], "output_channels")
        ):
            output_hash["output_channels"] = self.layers[-1].output_channels
            output_hash["output_height"] = self.layers[-1].output_height
            output_hash["output_width"] = self.layers[-1].output_width

            template = (
                Path(self.template_path)
                / "memory_layout/template_channels_last_output.c.tpl"
            ).read_text()
            mustach_hash["ouput_str"] = pystache.render(template, output_hash)
        else:
            output_hash["output_size"] = self.layers[-1].size
            if self.data_format == "channels_first":
                output_hash["comment"] = (
                    "Returning the output in channels first (ACETONE compute the result in channels first)"
                )
            else:
                output_hash["comment"] = "Returning the output (output flatten)"

            template = (
                Path(self.template_path)
                / "memory_layout/template_channels_first_output.c.tpl"
            ).read_text()
            mustach_hash["ouput_str"] = pystache.render(template, output_hash)

        if self.normalize:
            mustach_hash["pre_processing"] = self.Normalizer.write_pre_processing()
            mustach_hash["post_processing"] = self.Normalizer.write_post_processing()

        # Generate C code
        template = (Path(self.template_path) / "template_source_file.c.tpl").read_text()
        with (output_dir / "inference.c").open("a+") as source_file:
            source_file.write(pystache.render(template, mustach_hash))

    def generate_function_header_file(self, output_dir: Path):
        """Generate C Code for graph structure."""
        mustach_hash = {
            "data_type": self.data_type,
            "path": [i for i in range(self.maxpath)],
        }

        self.nb_weights_max = 1
        self.nb_biases_max = 1

        self.patches_size_max = 1
        self.concate_size_max = 0
        for layer in self.layers:
            if (
                isinstance(layer, Conv2D_std_gemm)
                and layer.patches_size > self.patches_size_max
            ):
                self.patches_size_max = layer.patches_size
            if isinstance(layer, Concatenate):
                self.patches_size_max = max(self.patches_size_max, layer.size)

        if any(isinstance(layer, Conv2D_std_gemm) for layer in self.layers):
            mustach_hash["path_size"] = max(self.l_size_max, self.patches_size_max)
        else:
            mustach_hash["path_size"] = self.l_size_max

        mustach_hash["cst"] = []
        written = {}
        for layer in self.dict_cst:
            if self.dict_cst[layer.idx] not in written:
                written[self.dict_cst[layer.idx]] = layer.size
            else:
                written[self.dict_cst[layer.idx]] = max(
                    written[self.dict_cst[layer.idx]],
                    layer.size,
                )

        for cst in written:
            mustach_hash["cst"].append({"name": cst, "size": written[cst]})

        if (
            any(
                isinstance(layer, Concatenate)
                or any(isinstance(layer, Conv2D))
                or any(isinstance(layer, Dense))
                or any(issubclass(layer, Broadcast))
                or any(isinstance(layer, Gather))
                or any(isinstance(layer, Pad)),
            )
            for layer in self.layers
        ):
            mustach_hash["temp_size"] = max(self.l_size_max, self.patches_size_max)

        # Collect layer parameters
        mustach_hash["layers"] = []
        for layer in self.layers:
            to_print = False
            layer_hash = {"name": layer.name, "idx": f"{layer.idx:02d}"}

            if hasattr(layer, "weights"):
                layer_hash["nb_weights"] = layer.nb_weights
                if layer.nb_weights > self.nb_weights_max:
                    self.nb_weights_max = layer.nb_weights
                to_print = True

            if hasattr(layer, "biases"):
                layer_hash["nb_biases"] = layer.nb_biases
                if layer.nb_biases > self.nb_biases_max:
                    self.nb_biases_max = layer.nb_biases
                to_print = True

            if isinstance(layer, Conv2D_indirect_gemm):
                layer_hash["patches_size"] = layer.patches_size
                to_print = True

            if isinstance(layer, BatchNormalization):
                layer_hash["channels"] = layer.output_channels
                to_print = True

            if issubclass(type(layer), Broadcast):
                if layer.constant is not None:
                    layer_hash["constant_size"] = layer.constant_size
                    to_print = True

            if to_print:
                mustach_hash["layers"].append(layer_hash)

        if self.normalize:
            mustach_hash["normalization_cst"] = (
                self.Normalizer.write_normalization_cst_in_header_file()
            )

        # Generate header code
        template = (Path(self.template_path) / "template_header_file.c.tpl").read_text()
        with (output_dir / "inference.h").open("a+") as header_file:
            header_file.write(pystache.render(template, mustach_hash))

    def generate_globalvars_file(self, output_dir: Path):
        """Generate C Code for layer data."""
        mustach_hash = {
            "data_type": self.data_type,
            "path": [i for i in range(self.maxpath)],
        }

        if any(isinstance(layer, Conv2D_std_gemm) for layer in self.layers):
            mustach_hash["path_size"] = max(self.l_size_max, self.patches_size_max)
        else:
            mustach_hash["path_size"] = self.l_size_max

        mustach_hash["cst"] = []
        written = {}
        for layer in self.dict_cst:
            if self.dict_cst[layer.idx] not in written:
                written[self.dict_cst[layer.idx]] = layer.size
            else:
                written[self.dict_cst[layer.idx]] = max(
                    written[self.dict_cst[layer.idx]],
                    layer.size,
                )

        for cst in written:
            mustach_hash["cst"].append({"name": cst, "size": written[cst]})

        if (
            any(
                isinstance(layer, Concatenate)
                or any(isinstance(layer, Conv2D))
                or any(isinstance(layer, Dense))
                or any(issubclass(layer, Broadcast))
                or any(isinstance(layer, Gather))
                or any(isinstance(layer, Pad)),
            )
            for layer in self.layers
        ):
            mustach_hash["temp_size"] = max(self.l_size_max, self.patches_size_max)

        if any(isinstance(layer, Conv2D_indirect_gemm) for layer in self.layers):
            mustach_hash["zero"] = True

        mustach_hash["layers"] = []

        for layer in self.layers:
            to_print = False
            layer_hash = {"name": layer.name, "idx": f"{layer.idx:02d}"}

            if hasattr(layer, "weights"):
                layer_hash["nb_weights"] = layer.nb_weights
                layer_hash["weights"] = self.flatten_array(layer.weights)
                to_print = True

            if hasattr(layer, "biases"):
                layer_hash["nb_biases"] = layer.nb_biases
                layer_hash["biases"] = self.flatten_array_orderc(layer.biases)
                to_print = True

            if type(layer) is Conv2D_indirect_gemm:
                layer_hash["patches_size"] = layer.patches_size
                layer_hash["patches"] = layer.create_ppatches()
                to_print = True

            if type(layer) is BatchNormalization:
                layer_hash["channels"] = layer.output_channels
                layer_hash["mean"] = self.flatten_array_orderc(layer.mean)
                layer_hash["var"] = self.flatten_array_orderc(layer.var)
                layer_hash["scale"] = self.flatten_array_orderc(layer.scale)
                to_print = True

            if issubclass(type(layer), Broadcast):
                if layer.constant is not None:
                    layer_hash["constant"] = self.flatten_array_orderc(layer.constant)
                    layer_hash["constant_size"] = layer.constant_size
                    to_print = True

            if to_print:
                mustach_hash["layers"].append(layer_hash)

        template = Path(
            self.template_path + "template_global_var_file.c.tpl",
        ).read_text()
        with (output_dir / "global_vars.c").open("a+") as globalvars_file:
            globalvars_file.write(pystache.render(template, mustach_hash))
            if self.normalize:
                self.globalvars_file.write(
                    self.Normalizer.write_normalization_cst_in_globalvars_file(),
                )



"""
 *******************************************************************************
 * ACETONE: Predictable programming framework for ML applications in safety-critical systems
 * Copyright (c) 2022. ONERA
 * This file is part of ACETONE
 *
 * ACETONE is free software ;
 * you can redistribute it and/or modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation ;
 * either version 3 of  the License, or (at your option) any later version.
 *
 * ACETONE is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY ;
 * without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License along with this program ;
 * if not, write to the Free Software Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307  USA
 ******************************************************************************
"""

from .Pad import Pad

import pystache

#The Constant mode of the Pad layers
#Use a constant to fill paddings
class Constant_Pad(Pad):
    
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.mode = 'constant'
    
    def generate_inference_code_layer(self):

        output_str = self.previous_layer[0].output_str

        mustach_hash = {}

        mustach_hash['name'] = self.name
        mustach_hash['idx'] = "{:02d}".format(self.idx)
        mustach_hash['comment'] = self.activation_function.comment
        mustach_hash['size'] = self.size
        mustach_hash['output_str'] = output_str
        mustach_hash['road'] = self.path

        mustach_hash['activation_function'] = self.activation_function.write_activation_str('tensor_temp[j + ' + str(self.output_width) + ' * (i + ' + str(self.output_height) + ' * f)]')

        mustach_hash['output_channels'] = self.output_channels
        mustach_hash['output_height'] = self.output_height
        mustach_hash['output_width'] = self.output_width
        mustach_hash['pads_front'] = self.pads[1]
        mustach_hash['channels_without_pads_back'] = self.output_channels - self.pads[5]
        mustach_hash['pads_top'] = self.pads[2]
        mustach_hash['height_without_pads_bottom'] = self.output_height - self.pads[6]
        mustach_hash['pads_left'] = self.pads[3]
        mustach_hash['width_without_pads_right'] = self.output_width - self.pads[7]
        mustach_hash['constant'] = self.constant_value
        mustach_hash['input_width'] = self.input_shape[3]
        mustach_hash['input_height'] = self.input_shape[2]

        if(self.activation_function.name == 'linear'):
            mustach_hash['linear'] = True
        
        if(self.fused_layer):
            mustach_hash['fused_layer'] = self.fused_layer.write_activation_str('tenser_temp[j + ' + str(self.output_width) + ' * (i + ' + str(self.output_height) + ' * f)]')

        with open(self.template_path+'layers/Pad/template_Constant_Pad.c.tpl','r') as template_file:
            template = template_file.read()
        template_file.close()

        return pystache.render(template, mustach_hash)




"""
 *******************************************************************************
 * ACETONE: Predictable programming framework for ML applications in safety-critical systems
 * Copyright (c) 2022. ONERA
 * This file is part of ACETONE
 *
 * ACETONE is free software ;
 * you can redistribute it and/or modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation ;
 * either version 3 of  the License, or (at your option) any later version.
 *
 * ACETONE is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY ;
 * without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License along with this program ;
 * if not, write to the Free Software Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307  USA
 ******************************************************************************
"""

import argparse
import pathlib

from .code_generator.neural_network import CodeGenerator

def cli_acetone(model_file:str, function_name:str, nb_tests:int, conv_algorithm:str, output_dir:str, test_dataset_file:str|None=None, normalize:bool=False ):
    print("C CODE GENERATOR FOR NEURAL NETWORKS")

    pathlib.Path(output_dir).mkdir(parents=True, exist_ok=True)

    net = CodeGenerator(file = model_file, test_dataset = test_dataset_file, function_name = function_name, nb_tests = nb_tests, conv_algorithm = conv_algorithm, normalize = normalize)
    net.generate_c_files(output_dir)
    net.compute_inference(output_dir)

    
if __name__ == "__main__":

    parser = argparse.ArgumentParser(description='C code generator for neural networks')

    parser.add_argument("model_file", help="Input file that describes the neural network model")
    parser.add_argument("function_name", help="Name of the generated function")
    parser.add_argument("nb_tests", help="Number of inferences process to run")
    parser.add_argument("conv_algorithm", help="Algorithm to be used in convolutional layer. Default is indirect im2col with GeMM")
    parser.add_argument("output_dir", help="Output directory where generated files will be written")
    parser.add_argument("test_dataset_file", nargs='?', default=None, help="Input file that contains test data")
    parser.add_argument("normalize", nargs='?', default=False, help="Boolean saying if the inputs and outputs needs to be normalized. Only used when the file is in NNET representation")

    args = parser.parse_args()

    cli_acetone(args.model_file, args.function_name, args.nb_tests, args.conv_algorithm, args.output_dir, args.test_dataset_file, args.normalize)



"""
 *******************************************************************************
 * ACETONE: Predictable programming framework for ML applications in safety-critical systems
 * Copyright (c) 2022. ONERA
 * This file is part of ACETONE
 *
 * ACETONE is free software ;
 * you can redistribute it and/or modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation ;
 * either version 3 of  the License, or (at your option) any later version.
 *
 * ACETONE is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY ;
 * without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License along with this program ;
 * if not, write to the Free Software Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307  USA
 ******************************************************************************
"""

from ..Layer import Layer
from ..activation_functions import ActivationFunctions
import numpy as np
import pystache

#extract a list of subtensor from a given tensor
#attribut: axis alongside of which the submatrix will be extracted (if the desired submatrix must have the height, width or channels of the parent tensor)
#input: a tensor
#output: a list of tensor
class GatherElements(Layer):
    
    def __init__(self, idx:int, size:int, axis:int,  indices:np.ndarray, input_shape:list, output_shape:list, activation_function:ActivationFunctions):
        
        super().__init__()
        self.idx = idx
        self.size = size
        self.name = 'GatherElements'
        self.indices = indices
        self.axis = axis
        self.input_channels = input_shape[1]
        self.input_height = input_shape[2]
        self.input_width = input_shape[3]
        self.output_channels = output_shape[1]
        self.output_height = output_shape[2]
        self.output_width = output_shape[3]
        self.activation_function = activation_function

        ####### Checking the instantiation#######

        ### Checking argument type ###
        if  type(self.idx)!= int:
            raise TypeError("Error: idx type in GatherElements (idx must be int)")
        if  type(self.size)!= int:
            raise TypeError("Error: size type in GatherElements (size must be int)")
        if type(self.axis) != int:
            raise TypeError("Erro: axis type in GatherElements (axis must be int)")
        if type(self.indices) != np.ndarray:
            raise TypeError("Error: indices in GatherElements (indices must be an numpy array)")
        if type(self.output_channels) != int:
            raise TypeError("Error: output channels type in GatherElements (must be int)")
        if type(self.output_height) != int:
            raise TypeError("Error: output height type in GatherElements (must be int)")
        if type(self.output_width) != int:
            raise TypeError("Error: output width type in GatherElements (must be int)")
        if type(self.input_channels) != int:
            raise TypeError("Error: input channels type in GatherElements (must be int)")
        if type(self.input_height) != int:
            raise TypeError("Error: input height type in GatherElements (must be int)")
        if type(self.input_width) != int:
            raise TypeError("Error: input width type in GatherElements (must be int)")
        if not isinstance(self.activation_function, ActivationFunctions):
            raise TypeError("Error: activation function type in GatherElements (activation function must be a sub-classe of acetone_nnet Activation Function)")

        ### Checking value consistency ###
        if self.size != self.output_channels*self.output_height*self.output_width:
            raise ValueError("Error: size value in GatherElements ("+str(self.size)+"!="+str(self.output_channels*self.output_height*self.output_width)+")")
        if axis not in [1,2,3]:
            raise ValueError("Error: axis out of bound in GatherElements ("+str(axis)+"for tensor in 4 dimension with first dimension unused)")
        for indice in self.indices.flatten():
            if indice < 0 or indice >= input_shape[self.axis]:
                raise ValueError("Error: indice out of bound in GatherElements ("+str(indice)+"out of bound for input of size"+str(input_shape[self.axis])+")")
        if self.indices.shape[1:] != (self.output_channels, self.output_height, self.output_width):
            raise ValueError("Error: non consistency between the indice shape and the output shape in GatherElements ("+str(self.indices.shape[1:])+"!="+str((self.output_channels, self.output_height, self.output_width))+")")
        if self.axis != 1 and self.output_channels != self.input_channels:
            raise ValueError("Error: non consistency between the input shape and the output shape in GatherElements ("+str(input_shape)+"!="+str((self.output_channels,self.output_height,self.output_width))+")")
        if self.axis != 2 and self.output_height != self.input_height:
            raise ValueError("Error: non consistency between the input shape and the output shape in GatherElements ("+str(input_shape)+"!="+str((self.output_channels,self.output_height,self.output_width))+")")
        if self.axis != 3 and self.output_width != self.input_width:
            raise ValueError("Error: non consistency between the input shape and the output shape in GatherElements ("+str(input_shape)+"!="+str((self.output_channels,self.output_height,self.output_width))+")")
        
    def generate_inference_code_layer(self):
        output_str = self.previous_layer[0].output_str

        mustach_hash = {}

        mustach_hash['name'] = self.name
        mustach_hash['idx'] = "{:02d}".format(self.idx)
        mustach_hash['comment'] = self.activation_function.comment
        mustach_hash['output_str'] = output_str
        mustach_hash['road'] = self.path
        mustach_hash['size'] = self.size

        mustach_hash['activation_function'] = self.activation_function.write_activation_str('tensor_temp[k]')

        mustach_hash['input_width'] = self.input_width
        mustach_hash['input_height'] = self.input_height
        mustach_hash['output_channels'] = self.output_channels
        mustach_hash['output_height'] = self.output_height
        mustach_hash['output_width'] = self.output_width

        if(self.axis == 1):
            mustach_hash['channels'] = True
        elif(self.axis == 2):
            mustach_hash['heights'] = True
        elif(self.axis == 3):
            mustach_hash['widths'] = True

        with open(self.template_path+'layers/template_GatherElements.c.tpl','r') as template_file:
            template = template_file.read()
        template_file.close()

        return pystache.render(template, mustach_hash)
        
    def forward_path_layer(self, input:np.ndarray):
        input = input.reshape(self.input_channels,self.input_height,self.input_width)
        output = np.zeros((self.output_channels, self.output_height, self.output_width))
        for f in range(self.output_channels):
            for i in range(self.output_height):
                for j in range(self.output_width):
                    if self.axis == 1:
                        output[f,i,j] = input[self.indices[0,f,i,j],i,j]
                    elif self.axis == 2:
                        output[f,i,j] = input[f,self.indices[0,f,i,j],j]
                    elif self.axis == 3:
                        output[f,i,j] = input[f,i,self.indices[0,f,i,j]]
        return output



"""
 *******************************************************************************
 * ACETONE: Predictable programming framework for ML applications in safety-critical systems
 * Copyright (c) 2022. ONERA
 * This file is part of ACETONE
 *
 * ACETONE is free software ;
 * you can redistribute it and/or modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation ;
 * either version 3 of  the License, or (at your option) any later version.
 *
 * ACETONE is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY ;
 * without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License along with this program ;
 * if not, write to the Free Software Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307  USA
 ******************************************************************************
"""

import pystache
import numpy as np
from abc import ABC
from ... import templates

class Normalizer(ABC):

    def __init__(self, input_size:int, output_size:int, mins:list, maxes:list, means:list, ranges:list):
        self.input_size = input_size
        self.output_size = output_size
        self.mins = mins
        self.maxes = maxes
        self.means = means
        self.ranges = ranges
        self.template_path = templates.__file__[:-11]
        super().__init__()

    def array_to_str(self, array:list):
        s = "{"
        for element in array:
            s += str(element) + ", "
        s = s[:-2] + "}"
        return s

    def write_pre_processing(self):
        with open(self.template_path+'normalization/template_pre_processing.c.tpl','r') as template_file:
            template = template_file.read()
        template_file.close()

        return pystache.render(template,{'input_size':self.input_size})


    def write_post_processing(self):
        with open(self.template_path+'normalization/template_post_processing.c.tpl','r') as template_file:
            template = template_file.read()
        template_file.close()

        return pystache.render(template,{'output_size':self.output_size})

    def write_normalization_cst_in_header_file(self):
        mustach_hash= {}

        mustach_hash['input_size'] = self.input_size

        with open(self.template_path+'normalization/template_normalization_cst_in_header_file.c.tpl','r') as template_file:
            template = template_file.read()
        template_file.close()

        return pystache.render(template, mustach_hash)
    
    def write_normalization_cst_in_globalvars_file(self):
        mustach_hash= {}

        mustach_hash['input_size'] = self.input_size
        mustach_hash['input_min'] = self.array_to_str(self.mins)
        mustach_hash['input_max'] = self.array_to_str(self.maxes)
        mustach_hash['input_mean'] = self.array_to_str(self.means[:-1])
        mustach_hash['input_range'] = self.array_to_str(self.ranges[:-1])
        mustach_hash['output_mean'] = self.means[-1]
        mustach_hash['output_range'] = self.ranges[-1]

        with open(self.template_path+'normalization/template_normalization_cst_in_global_var_file.c.tpl','r') as template_file:
            template = template_file.read()
        template_file.close()

        return  pystache.render(template, mustach_hash)

    def pre_processing(self, nn_input:np.ndarray):
        inputs = nn_input.flatten()
        for i in range(self.input_size):
            inputs[i] = (max(self.mins[i],min(self.maxes[i],inputs[i])) - self.means[i]) / self.ranges[i]
        return np.reshape(inputs, nn_input.shape)

    def post_processing(self, nn_output:np.ndarray):
        for i in range(len(nn_output)):
            nn_output[i] = nn_output[i]*self.ranges[-1] + self.means[-1]
        return nn_output


"""Base Layer type definition.

*******************************************************************************
* ACETONE: Predictable programming framework for ML applications in safety-critical systems
* Copyright (c) 2022. ONERA
* This file is part of ACETONE
*
* ACETONE is free software ;
* you can redistribute it and/or modify it under the terms of the GNU Lesser General Public
* License as published by the Free Software Foundation ;
* either version 3 of  the License, or (at your option) any later version.
*
* ACETONE is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY ;
* without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
* See the GNU Lesser General Public License for more details.
*
* You should have received a copy of the GNU Lesser General Public License along with this program ;
* if not, write to the Free Software Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307  USA
******************************************************************************
"""

from abc import ABC, abstractmethod
from dataclasses import dataclass
from pathlib import Path

import numpy as np
from typing_extensions import Self

from acetone_nnet import templates


@dataclass
class Layer(ABC):
    """Base class for inference layer."""

    def __init__(self: Self) -> None:
        """Build a non-specific layer."""
        self.idx = 0
        self.size = 0
        self.name = ""
        self.next_layer: list[Layer] = []
        self.previous_layer: list[Layer] = []
        self.path: int | None = None
        self.sorted: int | None = None
        self.output_str = ""
        self.fused_layer = None
        self.template_path = (
            str(
                Path(templates.__file__).parent,
            )
            + "/"
        )  # FIXME Should be a proper path

        super().__init__()

    @abstractmethod
    def generate_inference_code_layer(self: Self) -> str:
        """Generate computation code for layer."""

    @abstractmethod
    def forward_path_layer(
        self: Self,
        inputs: np.ndarray | list[np.ndarray],
    ) -> np.ndarray:
        """Compute output of layer."""

    @staticmethod
    def count_elements_array(array: np.ndarray) -> int:
        """Count elements in numpy array."""
        return np.multiply.reduce(array=array.shape)

    @staticmethod
    def compute_padding(
        padding: str | list,
        in_height: int,
        in_width: int,
        kernel_h: int,
        kernel_w: int,
        strides: int,
        dilation_rate: int = 1,
    ) -> tuple[int, int, int, int]:
        """Compute required padding given input and kernel dimensions."""
        pad_right, pad_left, pad_bottom, pad_top = 0, 0, 0, 0
        if isinstance(padding, str):
            # Compute 'same' padding tensorflow

            filter_height = kernel_h - (kernel_h - 1) * (dilation_rate - 1)
            filter_width = kernel_w - (kernel_w - 1) * (dilation_rate - 1)

            # The total padding applied along the height and width is computed as:
            if padding in ("VALID", "valid"):
                pad_right, pad_left, pad_bottom, pad_top = 0, 0, 0, 0
            else:
                if in_height % strides == 0:
                    pad_along_height = max(filter_height - strides, 0)
                else:
                    pad_along_height = max(filter_height - (in_height % strides), 0)
                if in_width % strides == 0:
                    pad_along_width = max(filter_width - strides, 0)
                else:
                    pad_along_width = max(filter_width - (in_width % strides), 0)

                if padding in ("SAME_UPPER", "same"):
                    pad_top = pad_along_height // 2
                    pad_bottom = pad_along_height - pad_top
                    pad_left = pad_along_width // 2
                    pad_right = pad_along_width - pad_left
                elif padding == "SAME_LOWER":
                    pad_bottom = pad_along_height // 2
                    pad_top = pad_along_height - pad_bottom
                    pad_right = pad_along_width // 2
                    pad_left = pad_along_width - pad_right
        else:
            pad_right, pad_left, pad_bottom, pad_top = (
                padding[3],
                padding[1],
                padding[2],
                padding[0],
            )

        return pad_right, pad_left, pad_bottom, pad_top

    # Give to the layer an string saying were the output will be saved (either in a 'cst' or in an 'output_road')
    def find_output_str(self: Self, dict_cst: dict):
        # dict_cst is the dict linking an layer to the cst in which the must be saved if needed

        # either it has to be saved
        if len(dict_cst) and self in dict_cst:
            output_str = "cst_" + str(dict_cst[self])
        # Or it can directly go to the next layer
        else:
            output_str = "output_" + str(self.path)
        self.output_str = output_str
        return self
